#!/usr/bin/env python3

import sys
import yaml
from jinja2 import Template
import json
import argparse
import os
from shutil import copyfile
import requests
from pathlib import Path
from dna_workflows import wf_engine
from dna_workflows import schema_tools
from tabulate import tabulate

sys.path.append(os.getcwd())
module_file_path = './.modules'


def main(_args=None):

    # If we are being called from a test framework
    if _args is not None:
        args = parse_args(_args)
    else:
        args = parse_args(sys.argv[1:])

    if args.host:
        exec_wfaas(args)
    else:
        exec_wflocal(args)


def exec_wflocal(args):

    if args.job_status or args.job_files:
        print('args --job-status and --job-files require the --host param')
        return

    if args.build_xlsx:
        from dna_workflows import schema_tools
        from dna_workflows import package_tools
        modules = package_tools.build_module_list()
        wb = schema_tools.create_new_workbook()
        wb = schema_tools.build_module_schema(wb, modules, user_data={})
        wb = schema_tools.build_workflow_task_sheet(wb, modules)
        wb.save(args.build_xlsx)
        return

    if args.build_test_xlsx:
        from dna_workflows import schema_tools
        from dna_workflows import package_tools
        modules = package_tools.build_module_list()
        wb = schema_tools.create_new_workbook()
        wb = schema_tools.build_module_schema(wb, modules)
        wb = schema_tools.build_workflow_task_sheet(wb, modules)
        wb.save(args.build_test_xlsx)
        return

    if args.update_xlsx_schema:
        from dna_workflows import schema_tools
        from dna_workflows import package_tools
        new_file = args.update_xlsx_schema
        backup_file = 'save.{}'.format(args.update_xlsx_schema)
        try:
            copyfile(args.update_xlsx_schema, backup_file)
        except OSError as e:
            print('ERROR: could not create db backup file {}: {}'.format(backup_file, e))
            return
        modules = package_tools.build_module_list()
        _workflow_db = schema_tools.load_xl_wf_db(args.update_xlsx_schema, flatten=True, fill_empty=True)
        wb = schema_tools.create_new_workbook()
        wb = schema_tools.build_module_schema(wb, modules, user_data=_workflow_db)
        wb = schema_tools.build_workflow_task_sheet(wb, modules, user_data=_workflow_db)
        wb.save(new_file)
        return

    if args.install:
        from dna_workflows import package_tools
        return package_tools.install_manifest()

    if args.install_url:
        from dna_workflows import package_tools
        return package_tools.install_package_from_url(args.install_url)

    if args.install_zip:
        from dna_workflows import package_tools
        return package_tools.install_package_from_zip(args.install_zip)

    if args.add_module_skeleton:
        create_module_skeleton()
        return

    workflow_db = compile_workflow(args)
    if workflow_db is None:
        return None

    if args.validate:
        if args.db or args.yaml_db:
            from dna_workflows import schema_tools
            _result = schema_tools.validate_module_schema(workflow_db, stdout=True)
            return _result
        else:
            print('--validate requires that you specify a valid workflow DB via --db or --yaml-db')
            return None

    results = wf_engine.run_wf(workflow_db)
    print_workflow_results(results)

    return results


def exec_wfaas(args):

    workflow_db = compile_workflow(args)
    if workflow_db is None:
        return None

    if args.job_status:
        _job_id = args.job_status
        url = 'http://{}/job/status/{}'.format(workflow_db['options']['host'], _job_id)
        r = requests.get(url)
        job_status = r.json()
        print('\nWorkflow Report:')
        print(tabulate([
            ['Job ID:', job_status['id']],
            ['Start time:', job_status['details']['start_time']],
            ['End time:', job_status['details']['end_time']],
            ['Status:', job_status['details']['status']]
        ],
            tablefmt="pretty",
            colalign=("left", "left")
        ))
        if 'results' in job_status['details'].keys():
            print_workflow_results(job_status['details']['results'])

        return job_status

    elif args.job_files:
        _job_id = args.job_files
        url = 'http://{}/job/files/{}'.format(workflow_db['options']['host'], _job_id)
        r = requests.get(url)
        if r.status_code == 404:
            _error = r.json()
            print('ERROR: 404 Not found: {}'.format(_error['error']))
        elif r.status_code == 200:
            _save_file = 'results_{}.zip'.format(_job_id)
            file = open(_save_file, "wb")
            file.write(r.content)
            file.close()
            print('Saved job files to: {}'.format(_save_file))

    elif args.build_xlsx:
        _save_file = args.build_xlsx
        url = 'http://{}/schema/xlsx'.format(workflow_db['options']['host'])
        r = requests.get(url)
        if r.status_code == 404:
            _error = r.json()
            print('ERROR: 404 Not found: dna_workflows.xlsx: {}'.format(_error['error']))
        elif r.status_code == 200:
            file = open(_save_file, "wb")
            file.write(r.content)
            file.close()
            print('Saved schema file to: {}'.format(_save_file))

    elif args.build_test_xlsx:
        _save_file = args.build_test_xlsx
        url = 'http://{}/schema/xlsx'.format(workflow_db['options']['host'])
        r = requests.get(url)
        if r.status_code == 404:
            _error = r.json()
            print('ERROR: 404 Not found: dna_workflows.xlsx: {}'.format(_error['error']))
        elif r.status_code == 200:
            file = open(_save_file, "wb")
            file.write(r.content)
            file.close()
            print('Saved schema file to: {}'.format(_save_file))
    else:
        url = 'http://{}/workflow'.format(workflow_db['options']['host'])
        r = requests.post(url, data=json.dumps(workflow_db))
        job_urls = json.loads(r.json())

        print(tabulate([
            ['Job ID:', job_urls['id']],
            ['Status URL:', job_urls['job_status']],
            ['Files URL:', job_urls['job_files']]
        ],
            headers=['Item', 'URL'],
            tablefmt="pretty",
            colalign=("left", "left")
        ))

    return "UNKNOWN"


def parse_args(args):
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--db", help=".xlsx file to use as the db")
    group.add_argument("--yaml-db", help=".yaml file to use as the db")
    parser.add_argument("--profile", help="Use this credentials profile")
    parser.add_argument("--build-xlsx", help="Builds a Excel workflow db based on the module manifest")
    parser.add_argument("--build-test-xlsx", help="Builds a Excel workflow db based on the module manifest with "
                                                  "prepopulated test data")
    # Needs review/fixing
    # parser.add_argument("--manifest", help="Used to specify a manifest file when building an xlsx schema.  Note that "
    #                                        "the modules must already be installed or available from the current "
    #                                        "working directory")
    parser.add_argument("--install", action='store_true', help="Install packages using a manifest from the current "
                                                               "working directory")
    parser.add_argument("--install-url", help="Install packages directly from URL. The URL must provide a "
                                              ".zip archive for the package.")
    parser.add_argument("--install-zip", help="Install packages directly from a .zip archive.")
    parser.add_argument("--update-xlsx-schema", help="Takes an existing Excel workflow DB and tries to update the "
                                                     "schema based on the latest module definition")
    parser.add_argument("--validate", action='store_true', help="Request DB schema schema validation")
    parser.add_argument("--noop", action='store_true', help="Run the scheduling logic but do not execute any workflows")
    parser.add_argument("--offline", action='store_true',
                        help="Creates a 'dummy' api object, useful for workflow development")
    parser.add_argument("--dump-db-to-yaml", help="Creates an yaml file from provided *.xlsx workbook")
    parser.add_argument("--debug", action='store_true', help="Enable debug level messages mode")
    parser.add_argument("--add-module-skeleton", action='store_true', help="Create a DNA Workflows module template")
    parser.add_argument("--host", help="Specify a host running the DNA Workflows Web App")
    parser.add_argument("--job-status", help="Retrieves the job status from DNAWFaaS given the job <id>.  Requires "
                                             "the --host argument.")
    parser.add_argument("--job-files", help="Retrieves the job files from DNAWFaaS given the job <id>.  Requires "
                                            "the --host argument.")

    return parser.parse_args(args)


def compile_workflow(args):

    # Load the workflow DB
    if args.db and args.yaml_db:
        print('WARNING: Only single source of data is allowed. YAML db file will be used.')
    if args.yaml_db:
        try:
            _workflow_db = yaml.load(open(args.yaml_db, 'r'), Loader=yaml.SafeLoader)
        except Exception as e:
            print('Unable to load YAML file {}'.format(args.yaml_db))
            print(e)
            return None
    elif args.db:
        try:
            _workflow_db = schema_tools.load_xl_wf_db(args.db, flatten=True)
        except Exception as e:
            print('Unable to load YAML file {}'.format(args.yaml_db))
            print(e)
            return None
    else:
        _workflow_db = {'workflow': [{'stage': '1', 'module': 'noop', 'task': 'noop', 'api': 'noop'}]}

    # Dump the workflow DB to Yaml
    if args.dump_db_to_yaml:
        with open(args.dump_db_to_yaml, 'w') as file:
            yaml.dump(_workflow_db, file)

    # Try to load credentials
    if args.offline:
        _workflow_db.update({'api_creds': {'offline': True}})
    else:
        if args.profile:
            api_creds = load_credentials(profile=args.profile)
        else:
            api_creds = load_credentials()

        _workflow_db.update({'api_creds': api_creds})

    # Add workflow options
    options = {}
    if args.debug: options.update({'logging': 'DEBUG'})
    if args.noop: options.update({'noop': True})
    if args.host: options.update({'host': args.host})

    _workflow_db.update({'options': options})

    return _workflow_db


def load_credentials(profile=None):
    _home_creds = '{}/.dna_workflows/credentials'.format(str(Path.home()))

    if os.path.isfile('./credentials'):
        _creds = yaml.load(open('./credentials', 'r'), Loader=yaml.SafeLoader)
        # return _creds
    elif os.path.isfile(_home_creds):
        _creds = yaml.load(open(_home_creds, 'r'), Loader=yaml.SafeLoader)
        # return _creds
    else:
        print('FATAL: Unable to find credentials in either ./credentials or ~/.dna_workflows/credentials')
        exit()

    if profile is not None:
        if profile in _creds.keys():
            _pcreds = _creds[profile]
            return _pcreds
        else:
            print('profile "{}" not found. Using default profile'.format(profile))
            return _creds
    else:
        return _creds


def create_module_skeleton():
    from dna_workflows import payload_templates
    _module = {}
    _module.update({'module': input('Module name: ')})
    _module.update({'email': input('email: ')})

    # Create module directory
    try:
        os.mkdir(_module['module'])
    except OSError as e:
        print('Creation of the directory {} failed: {}'.format(_module['module'], e))
    else:
        print('Successfully created the directory {}'.format(_module['module']))

    # Write module template files
    tm = Template(payload_templates.workflow_j2)
    output = tm.render(item=_module)
    write_to_file(_module['module'], 'workflow.py', output)

    tm = Template(payload_templates.module_j2)
    output = tm.render(item=_module)
    write_to_file(_module['module'], 'module', output)

    tm = Template(payload_templates.payload_templates_j2)
    output = tm.render()
    write_to_file(_module['module'], 'payload_templates.py', output)

    tm = Template(payload_templates.init_j2)
    output = tm.render()
    write_to_file(_module['module'], '__init__.py', output)

    _manifest = {'manifest': {_module['module']: {'type': 'module'}}}
    with open('manifest.yaml', 'w') as file:
        yaml.dump(_manifest, file)

    return


def write_to_file(path, file, data):
    _path_to_file = '{}/{}'.format(path, file)
    f = open(_path_to_file, "w")
    f.write(data)
    f.close()


def print_workflow_results(_results):
    _fresults = []
    for _task in _results:
        _fresults.append(
            ['Stage {}'.format(_task['stage']), _task['task'], _task['result']]
        )
    print('\nTask Report:')
    print(tabulate(
        _fresults,
        headers=['Stage', 'Task', 'Result'],
        tablefmt="pretty",
        colalign=("left", "left",)),
    )
    print('')


if __name__ == "__main__":
    main()



import sys
import traceback
import logging
from pathlib import Path
import os
import shutil
import yaml
import requests
import zipfile
import io
import copy

logger = logging.getLogger('package_tools')
home = str(Path.home())
install_dir = '{}/.dna_workflows/install'.format(home)


def install_manifest():
    Path(install_dir).mkdir(parents=True, exist_ok=True)

    _package_manifest = Path('manifest.yaml')
    if _package_manifest.is_file():
        try:
            _manifest = yaml.load(open(_package_manifest, 'r'), Loader=yaml.SafeLoader)
        except Exception as e:
            print('Unable to load manifest {}.  Please check this valid YAML'.format(_package_manifest))
            print(e)
            return
    else:
        print('ERROR: manifest file {} not found'.format(_package_manifest))
        return

    if 'manifest' not in _manifest.keys():
        logger.error('Error parsing manifest: BAD FORMAT {}'.format(_manifest))
        return -1

    # Copy the package files to the install_dir
    for _package_name, _package_meta in _manifest['manifest'].items():
        if os.path.isdir(_package_name):
            to_path = '{}/{}'.format(install_dir, _package_name)
            copy_and_overwrite(_package_name, to_path)
        else:
            logger.error('Could not find directory {}'.format(_package_name))
            return -1

    _install_manifest = load_install_manifest()
    _install_manifest['manifest'].update(_manifest['manifest'])

    # Write out install manifest
    _manifest_file = '{}/manifest.yaml'.format(install_dir)
    with open(_manifest_file, 'w') as file:
        yaml.dump(_install_manifest, file)


def load_install_manifest():
    _manifest_file = '{}/manifest.yaml'.format(install_dir)
    if os.path.isfile(_manifest_file):
        try:
            _manifest = yaml.load(open(_manifest_file, 'r'), Loader=yaml.SafeLoader)
        except Exception as e:
            print('Unable to load install manifest {}.  Please check this valid YAML'.format(_manifest_file))
            print(e)
            return
    else:
        _manifest = {'manifest': {}}

    return _manifest


def load_module(_module):
    try:
        sys.path.index(install_dir)
    except ValueError:
        sys.path.append(install_dir)

    _manifest = build_module_list()
    _import = 'import {}'.format(_module)

    if _module in globals():
        logger.debug('Module {} is already loaded'.format(_module))
    elif _module in _manifest:
        logger.debug('Loading module {}'.format(_module))
        exec(_import, globals())
    else:
        logger.error('Workflow module with name {} is not loaded'.format(_module))
        return -1

    return


def execute_task(api, _module, _task, _workflow_dict):
    _manifest = build_method_list(name_only_fully_qualified=True)
    _func = '{}.{}'.format(_module, _task)
    _task_exec = '{}(api, copy.deepcopy({}))'.format(_func, _workflow_dict)
    if _func in _manifest:
        result = eval(_task_exec)
    else:
        logger.error('Task {} from workflow module {} is not loaded'.format(_task, _module))
        return -1

    return result


def build_module_list():
    _install_manifest = load_install_manifest()
    _module_list = []
    for _package_name, _package_meta in _install_manifest['manifest'].items():
        if _package_meta['type'] == 'module':
            _module_list.append(_package_name)
        elif _package_meta['type'] == 'bundle':
            for _bundle_module in _package_meta['provides']:
                _module = '{}.{}'.format(_package_name, _bundle_module)
                _module_list.append(_module)

    return _module_list


def build_method_list(name_only=False, name_only_fully_qualified=False):
    _methods = []
    for _module in build_module_list():
        module_doc = get_module_doc(_module)

        for m in module_doc['module']['methods']:
            if name_only:
                _methods.append(m['task'])
            elif name_only_fully_qualified:
                _task = '{}.{}'.format(_module, m['task'])
                _methods.append(_task)
            else:
                _methods.append(m)

    return _methods


def get_module_doc(_module):
    load_module(_module)
    exec_str = '{}.get_module_definition()'.format(_module)
    _module_doc = eval(exec_str)

    return _module_doc


def install_package_from_zip(file):
    extract_path = './.package'
    Path(extract_path).mkdir(parents=True, exist_ok=True)

    try:
        z = zipfile.ZipFile(file)
        z.extractall(extract_path)
    except FileNotFoundError:
        logger.error('File not found: {}'.format(file))
        return -1

    file_list = os.listdir(extract_path)
    archive_root = '{}/{}'.format(extract_path, file_list[0])
    if len(file_list) == 1 and os.path.isdir(archive_root):
        logger.info('Archive extracted OK to path: {}'.format(archive_root))
        pass
    else:
        logger.error('Fatal error extracting package.  Archive should contain a single root dir: {}'.format(file_list))
        return -1

    # Get the current working dir, do the install then remove the download
    cwd = os.getcwd()
    os.chdir(archive_root)
    install_manifest()
    os.chdir(cwd)
    shutil.rmtree(extract_path)


def install_package_from_url(url):
    save_path = './package.zip'
    chunk_size = 128

    r = requests.get(url, stream=True)
    with open(save_path, 'wb') as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)

    install_package_from_zip(save_path)


def copy_and_overwrite(from_path, to_path):
    if os.path.exists(to_path):
        shutil.rmtree(to_path)
    shutil.copytree(from_path, to_path)


def transform_wf_tasks(workflow_db):
    _workflow_task_list = []

    # Collect workflow schemas
    for _schema_name, _schema_data in workflow_db.items():
        if 'workflow' in _schema_name:
            wf_tasks = reformat_wf_tasks(_schema_name, _schema_data)
            _workflow_task_list = _workflow_task_list + wf_tasks

    return _workflow_task_list


def reformat_wf_tasks(_schema_name, _schema_data):
    if _schema_name == 'workflow':
        # print('BACKWARD COMPAT')
        _wf_tasks = [_row for _row in _schema_data if _row.get('status', 'enabled') == 'enabled']
        return _wf_tasks
    elif len(_schema_name.split('.')) == 2 and 'workflow' in _schema_name:
        # print('NEW MODULE FORMAT')
        _wf_tasks = [_row for _row in _schema_data if _row.get('status', 'enabled') == 'enabled']
        return _wf_tasks
    elif len(_schema_name.split('.')) == 3 and 'workflow.bundle' in _schema_name:
        # print('NEW BUNDLE FORMAT')
        _bundle_name = _schema_name.split('.')[2]
        # print('Bundle name {}'.format(_bundle_name))
        _wf_tasks = [_row for _row in _schema_data if _row.get('status', 'enabled') == 'enabled']
        for _task in _wf_tasks:
            _task['module'] = '.'.join([_bundle_name, _task['module']])
        return _wf_tasks
    else:
        print('ERROR NO WORKFLOW SCHEMAS FOUND')
        return -1


module_j2 = '''
---
module:
  name: &module "{{ item.module }}"     # Module name
  author: &author "{{ item.email }}"    # Authors email
  # A description of the module will be inserted at the top of the Excel worksheet
  description: "This is default module description for {{ item.module }}"
  # A list of provided public functions (Tasks).  These are inserted into the 'workflow' table
  methods:
    - {"status": "enabled", "task": "hello_world", "stage": 1, "module": *module, "api": "dnacentersdk",
       "author": *author,
       "description": "Runs the hello world task!"
    }
  # A dictionary of table schemas describing the table column names.
  # These will appear on the '<module-name>' worksheet and will create basic data validations
  # Table names take the form <schema-name>.schema.<module-name>
  # see https://github.com/cunningr/sdtables
  schemas:
    hello_world:
      description: "Used by: hello_world"
      properties:
        presence:
          type: string
          enum: ['present', 'absent']
          default: 'present'
        key1:
          type: string
          enum: ['value1', 'value2', 'value3']
          default: 'value1'
        key2:
          type: boolean
          default: False
        example_tref:
          type: string
          tref: 'INDIRECT("another_table.schema.{{ item.module }}[testCol1]")'
        example_default:
          type: string
          default: '=hello_world.schema.{{ item.module }}[[#This Row],[key1]]&"/"&hello_world.schema.{{ item.module }}[[#This Row],[key2]]'
    another_table:
      description: "A test reference table"
      properties:
        testCol1:
          type: string
        testCol2:
          type: string
  # Example data to populate the above schema.
  # In addition the data is also validated against the schema
  data:
    hello_world:
      - {'presence': 'present', 'key1': 'value2', 'key2': True, 'example_tref': 'TWO'}
    another_table:
      - {'testCol1': 'ONE', 'testCol2': 'AAAA'}
      - {'testCol1': 'TWO', 'testCol2': 'BBBB'}
      - {'testCol1': 'THREE', 'testCol2': 'CCCC'}
'''

workflow_j2 = '''
import logging
from {{ item.module }} import payload_templates as templates
import yaml
import pkgutil

logger = logging.getLogger('main.{{ item.module }}')


def get_module_definition():
    data = pkgutil.get_data(__package__, 'module')
    return yaml.load(data, Loader=yaml.SafeLoader)


def hello_world(api, workflow_dict):
    """ Prints hello_world and some test data from example schema.

    :param api: An instance of the XYZ SDK class
    :param workflow_dict: A dictionary containing rows of example data (see module);

    :returns: Nothing """
    
    _schema = 'hello_world.schema.{{ item.module }}'
    logger.info('{{ item.module }}::hello_world')
    logger.debug('schema: {}'.format(_schema))

    print(_schema)
    if _schema in workflow_dict.keys():
        table_data = workflow_dict[_schema]

        for row in table_data:
            print(row)

'''

init_j2 = '''
"""
This is a python init file generated from template.  Replace this text with your workflow description
"""

__version__ = "0.0.1"

from .workflow import *
'''

payload_templates_j2 = '''
a_jinja_template_j2 = """
"aValue": {{ aParameter }}
"""
'''


import sdtables
from openpyxl import Workbook
from openpyxl.styles import Color, PatternFill, Font, Alignment
from openpyxl.styles.differential import DifferentialStyle
from openpyxl.formatting import Rule
from dna_workflows import package_tools

# manifest = 'manifest.yml'
# xl_out = 'example.xlsx'


def create_new_workbook():
    _wb = Workbook()
    _ws = _wb.active
    _wb.remove(_ws)

    return _wb


def load_xl_wf_db(excel_db, flatten=False, fill_empty=False):
    _data = sdtables.load_xl_db(excel_db, flatten=flatten)

    return _data


def build_module_schema(_wb, _modules, user_data=None):
    """ Builds Excel tables from schema based on module manifest.

    :param _wb: (object) An openpyxl workbook object
    :param modules: (dict) Dictionary describing the project modules loaded from schema.yml
    :param user_data: (dict) In the case that there is existing data to preserve we can provide the existing table data

    :returns: an updated openpyxl workbook object """

    for module in _modules:
        module_doc = package_tools.get_module_doc(module)

        if 'schemas' in module_doc['module']:
            ws = _wb.create_sheet(module_doc['module']['name'])
            ws.sheet_properties.tabColor = "009900"
        else:
            continue

        if 'description' in module_doc['module'].keys():
            ws.append({2: module_doc['module']['description']})
            ws['B1'].fill = PatternFill("solid", fgColor="b6e5eb")
            ws['B1'].alignment = Alignment(wrapText=True, vertical='top')
            ws.merge_cells(start_row=1, start_column=2, end_row=1, end_column=11)

        for module_schema in module_doc['module']['schemas'].keys():
            module_name = module_doc['module']['name']
            name = '{}.{}.{}'.format(module_schema, 'schema', module_name)
            schema_doc = module_doc['module']['schemas'][module_schema]
            if user_data is not None:
                if name in user_data.keys():
                    data = user_data[name]
                    sdtables.add_schema_table_to_worksheet(ws, name, schema_doc, data=data, table_style='TableStyleMedium2')
                else:
                    sdtables.add_schema_table_to_worksheet(ws, name, schema_doc, table_style='TableStyleMedium2')
            elif 'data' in module_doc['module']:
                if module_schema in module_doc['module']['data'].keys():
                    data = module_doc['module']['data'][module_schema]
                    sdtables.add_schema_table_to_worksheet(ws, name, schema_doc, data=data, table_style='TableStyleMedium2')
                else:
                    sdtables.add_schema_table_to_worksheet(ws, name, schema_doc, table_style='TableStyleMedium2')
            else:
                sdtables.add_schema_table_to_worksheet(ws, name, schema_doc, table_style='TableStyleMedium2')

    return _wb


def build_workflow_task_sheet(_wb, _modules, user_data=None):
    """ Builds the workflows cover sheet with a table containing available tasks
    based on the project manifest loaded from schema.yml.

    :param _wb: (object) An openpyxl workbook object
    :param _modules: (dict) Dictionary describing the project modules loaded from schema.yml
    :param user_data: (dict) In the case that there is existing data to preserve we can provide the existing table data

    :returns: an updated openpyxl workbook object """

    _manifest = package_tools.load_install_manifest()
    from dna_workflows import wf_engine
    wf_doc = wf_engine.get_module_definition()
    _ws = _wb.create_sheet("workflows", 0)
    _ws.sheet_properties.tabColor = "0080FF"

    for _package_name, _package_meta in _manifest['manifest'].items():
        _module_list = []
        if _package_meta['type'] == 'module':
            _module_list.append(_package_name)
            _schema_suffix = _package_name
        elif _package_meta['type'] == 'bundle':
            _schema_suffix = 'bundle.{}'.format(_package_name)
            for _bundle_module in _package_meta['provides']:
                _module = '{}.{}'.format(_package_name, _bundle_module)
                _module_list.append(_module)

        # Build a list of modules and tasks based on the manifest
        methods = []
        for _module in _module_list:
            module_doc = package_tools.get_module_doc(_module)

            for m in module_doc['module']['methods']:
                methods.append(m)

        _table_name = 'workflow.{}'.format(_schema_suffix)
        if user_data is not None:
            user_methods = get_data_from_schema(_table_name, user_data)
            for _m in methods:
                for row in user_methods:
                    if _m['module'] == row['module'] and _m['task'] == row['task']:
                        _m['status'] = row['status']

        wf_schema = wf_doc['module']['schemas']['workflow']
        sdtables.add_schema_table_to_worksheet(_ws, _table_name, wf_schema, data=methods, table_style='TableStyleLight14')

        # Add conditional formatting to workflow worksheet
        for table in _ws.tables.values():
            if _table_name == table.name:
                _tdef = table.ref
                red_fill = PatternFill(bgColor="9da19e")
                dxf = DifferentialStyle(fill=red_fill)
                r = Rule(type="expression", dxf=dxf, stopIfTrue=True)
                _formula = '${}="disabled"'.format(_tdef.split(':')[0])
                r.formula = [_formula]
                _ws.conditional_formatting.add(_tdef, r)

    return _wb


def get_data_from_schema(_schema_name, _db):
    if _schema_name in _db.keys():
        return _db[_schema_name]
    else:
        return []


def validate_module_schema(_workflow_db, stdout=False):
    """ Validates the pre-loaded workflow DB against the module schema.

    :param stdout: Print results to STDOUT
    :param _workflow_db: (object) A DNA Workflows db pre-loaded from either Excel or YAML

    :returns: Boolean depending on validation results """
    _workflow_db['workflow'] = package_tools.transform_wf_tasks(_workflow_db)
    _modules = []
    for _task in _workflow_db['workflow']:
        if 'enabled' in _task['status'] and _task['module'] not in _modules:
            _modules.append(_task['module'])

    for _module in _modules:
        module_doc = package_tools.get_module_doc(_module)
        module_name = module_doc['module']['name']
        print('VALIDATING module: {}'.format(module_name))
        if 'schemas' in module_doc['module']:
            for _schema, _properties in module_doc['module']['schemas'].items():
                _dnawf_schema_name = '{}.schema.{}'.format(_schema, module_name)
                if _dnawf_schema_name in _workflow_db.keys():
                    print('VALIDATING schema: {}'.format(_dnawf_schema_name))
                    _result = sdtables.validate_data(_properties, _workflow_db[_dnawf_schema_name])
                    if stdout:
                        print_validation_results(_result)
                    else:
                        return _result
                else:
                    print('scheam {} not found'.format(_dnawf_schema_name))

    return


def print_validation_results(_results):
    print('RESULT: {}'.format(_results['result']))
    for _r in _results['details']:
        if _r['result'] != 'OK':
            print('Row {} contains validation errors:'.format(_r['row']+1))
            print('------------------ERROR-----------------'.center(50))
            print(_r['result'])
            print('--------------------------------------'.center(50))

# if __name__ == "__main__":
#     schema = yaml.load(open(manifest, 'r'), Loader=yaml.SafeLoader)
#     wb = create_new_workbook()
#     wb = build_module_schema(wb, schema)
#     wb = build_workflow_task_sheet(wb, schema)
#     wb.save(xl_out)



from distutils.core import setup
setup(
    name='dna_workflows',
    packages=['dna_workflows'],
    package_data={
    'dna_workflows': ['module']
    },
    version='0.0.21',
    license='MIT',
    description='dna_workflows is a basic workflow engine for executing DNA Workflows packages',
    author='Richard Cunningham',
    author_email='cunningr@gmail.com',
    url='https://github.com/cunningr/dna_workflows',
    download_url='https://github.com/cunningr/dna_workflows',
    keywords=['DNA Center', 'Cisco', 'workflow'],
    install_requires=[
      'jinja2',
      'ipaddress',
      'argparse',
      'requests',
      'netaddr',
      'pyyaml',
      'coloredlogs',
      'tabulate',
      'sdtables>=1.0.8',
      'dnacentersdk',
      'ISE'
    ],
    classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6'
    ],
    # scripts=['bin/dna_workflows'],
    entry_points={
        'console_scripts': [
            'dna_workflows = dna_workflows.wf_client:main'
        ]
    }
)



import yaml
import traceback
import copy
import urllib3
import logging
import coloredlogs
import json
import sys
import pkgutil
from dnacentersdk import DNACenterAPI
from dna_workflows import package_tools
from ise import ERS

# Settings
urllib3.disable_warnings()
logger = logging.getLogger('main')
module_file_path = './.modules'


def get_module_definition():
    data = pkgutil.get_data(__package__, 'module')
    return yaml.load(data, Loader=yaml.SafeLoader)


def run_wf(_workflow_db, headless=False):
    apis = run_setup(_workflow_db, headless=headless)

    wf_tasks = _workflow_db['workflow']

    execution_schedule = build_workflow_schedule(wf_tasks)
    _report = []
    for _task in execution_schedule:
        _task_api = _task[3]
        _mod_task = '{}.{}'.format(_task[1], _task[2])
        if _task_api in apis.keys():
            api = apis[_task_api]
            try:
                _result = execute_task(_task, api, _workflow_db)
                # _mod_task = '{}.{}'.format(_task[1], _task[2])
                if _result in ['FAILURE', 'SUCCESS', 'ERROR', 'PARTIAL_FAILURE', 'NOOP']:
                    _report.append({'stage': _task[0], 'task': _mod_task, 'result': _result})
                else:
                    _report.append({'stage': _task[0], 'task': _mod_task, 'result': 'UNKNOWN'})
            except Exception as e:
                logger.error('TASK EXCEPTION: API {}, Task {}'.format(_task_api, _task))
                logger.error('**** TRACEBACK ***\n\n {}'.format(traceback.format_exc()))
        # elif 'noop' in _task_api:
        #     api = 'noop'
        #     execute_task(_task, api, _workflow_db)
        elif 'offline' in apis.keys():
            api = 'offline'
            _result = execute_task(_task, api, _workflow_db)
            _report.append({'stage': 0, 'task': _mod_task, 'result': _result})
        else:
            logger.error('api: {} not found.  Please check your credentials file'.format(_task_api))
            exit()

    return _report


def build_workflow_schedule(wf_tasks):
    _func_tuple_list = []

    for _func in wf_tasks:
        if 'task' in _func.keys():
            _func_tuple = (_func['stage'], _func['module'], _func['task'], _func['api'])

        _func_tuple_list.append(_func_tuple)

    return sorted(_func_tuple_list, key=lambda tup: tup[0])


def execute_task(_task, api, _workflow_db):

    _stage = _task[0]
    _module = _task[1]
    _workflow_dict = _workflow_db
    _task = _task[2]

    if 'options' in _workflow_db.keys():
        options = _workflow_db['options']
    else:
        options = {}

    if 'noop' in _workflow_db['options'] or 'offline' == api:
        logger.info('Executing STAGE-{} workflow: {}::{}'.format(_stage, _module, _task))
        return 'NOOP'
    else:

        logger.info('Executing STAGE-{} workflow: {}::{}'.format(_stage, _module, _task))

        return package_tools.execute_task(api, _module, _task, _workflow_dict)


def run_setup(_workflow_db, headless=True):
    global logger

    if 'options' in _workflow_db.keys():
        options = _workflow_db['options']
    else:
        options = {}

    # Setup logging
    if 'logging' in options.keys():
        user_level = options['logging'].upper()
        level = logging.getLevelName(user_level)
    else:
        level = logging.getLevelName('INFO')

    coloredlogs.install(level=level, logger=logger, fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                        field_styles={'asctime': {'color': 'green'}, 'hostname': {'color': 'magenta'},
                                        'levelname': {'bold': True, 'color': 'black'}, 'name': {'color': 'yellow'},
                                        'programname': {'color': 'cyan'}, 'username': {'color': 'yellow'}})

    if headless:
        logger.info('DNA Workflow running in HEADLESS mode')
        fh = logging.FileHandler('workflow.log')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        fh.setLevel(level)
        logging.getLogger().addHandler(fh)

    api = {}
    if 'offline' in _workflow_db['api_creds'].keys():
        api = {'offline': True}
    else:
        nosdk = True
        if 'dnacentersdk' in _workflow_db['api_creds'].keys():
            _sdk = sdk_setup_dnacentersdk(_workflow_db['api_creds'])
            api.update({'dnacentersdk': _sdk})
            nosdk = False
        if 'isepac' in _workflow_db['api_creds'].keys():
            _sdk = sdk_setup_isepac(_workflow_db['api_creds'])
            api.update({'isepac': _sdk})
            nosdk = False

        # If we didn't find an SDK then exit
        if nosdk:
            logger.error('No valid SDK credentials found')
            exit()

    _workflow_db['workflow'] = package_tools.transform_wf_tasks(_workflow_db)

    return api


def get_options(_workflow_db, option):

    if 'options' in _workflow_db.keys():
        options = _workflow_db['options']
    else:
        return None

    if option in options.keys():
        return options[option]
    else:
        return None


def sdk_setup_dnacentersdk(api_creds):
    try:
        username = api_creds['dnacentersdk']['username']
        password = api_creds['dnacentersdk']['password']
        base_url = api_creds['dnacentersdk']['base_url']
        version = api_creds['dnacentersdk']['api_version']
        verify = str(api_creds['dnacentersdk']['verify']).lower() in ['true']
        api = DNACenterAPI(base_url=base_url, version=version, username=username, password=password, verify=verify)
        logger.info('API connectivity established with dnacentersdk')
        return api
    except Exception as e:
        logger.error('error connecting to dnacentersdk.  Please verify connectivity, username and password')
        logger.error(e)
        exit()


def sdk_setup_isepac(api_creds):
    try:
        username = api_creds['isepac']['username']
        password = api_creds['isepac']['password']
        host = api_creds['isepac']['host']
        verify = str(api_creds['isepac']['verify']).lower() in ['true']
        disable_warnings = str(api_creds['isepac']['disable_warnings']).lower() in ['true']
        api = ERS(ise_node=host, ers_user=username, ers_pass=password, verify=verify, disable_warnings=disable_warnings)
        logger.info('API connectivity established with isepac')
        return api
    except Exception as e:
        logger.error('error connecting to isepac.  Please verify connectivity, username and password')
        logger.error(e)
        exit()
